"""Module to test Account class."""

import pytest

from mailerlite.account import Account
from mailerlite.constants import API_KEY_TEST


@pytest.fixture
def header():
    headers = {'content-type': "application/json",
               'x-mailerlite-apikey': API_KEY_TEST
               }
    return headers


def test_wrong_headers():
    headers_2 = {'content-type': "application/json",
                 'x-mailerlite-apikey': 'FAKE_KEY'
                 }
    headers_3 = {'content-type': "application/json",
                 }
    headers_4 = {'x-mailerlite-apikey': 'FAKE_KEY'
                 }

    with pytest.raises(OSError):
        acc = Account(headers_2)
        acc.info()

    with pytest.raises(ValueError):
        acc = Account(headers_3)

    with pytest.raises(ValueError):
        acc = Account(headers_4)


def test_account(header):
    acc = Account(header)

    info = acc.info()
    assert 'account' in info.keys()

    stats = acc.stats()
    assert 'subscribed' in stats.keys()
    assert 'unsubscribed' in stats.keys()
    assert stats.get('subscribed') > stats.get('unsubscribed')

    optin = acc.double_optin()
    assert 'enabled' in optin.keys()

    with pytest.raises(OSError):
        # double_optin not available with this API keys
        acc.set_double_optin(False)



"""Module to tests Campaign."""

import pytest

from mailerlite.campaign import Campaigns
from mailerlite.constants import API_KEY_TEST


@pytest.fixture
def header():
    headers = {'content-type': "application/json",
               "X-MailerLite-ApiDocs": "true",
               'x-mailerlite-apikey': API_KEY_TEST
               }
    return headers


@pytest.fixture
def campaign_data():
    data = {"subject": "Regular campaign subject",
            "name": "Regular campaign name",
            "groups": [2984475, 3237221],
            "type": "regular"
            }
    return data


@pytest.fixture
def campaign_data_ab():
    data_ab = {"groups": [2984475, 3237221],
               "type": "ab",
               "ab_settings": {"send_type": "subject",
                               "values": ["Email subject A",
                                          "Email subject B"],
                               "ab_win_type": "opens",
                               "winner_after": 1,
                               "winner_after_type": "h",
                               "split_part": "10"
                               }
               }
    return data_ab


def test_wrong_headers(campaign_data):
    headers_1 = {'content-type': "app",
                 "X-MailerLite-ApiDocs": "true",
                 'x-mailerlite-apikey': API_KEY_TEST
                 }
    headers_2 = {'content-type': "application/json",
                 "X-MailerLite-ApiDocs": "true",
                 'x-mailerlite-apikey': 'FAKE_KEY'
                 }
    headers_3 = {'content-type': "application/json",
                 }
    headers_4 = {'x-mailerlite-apikey': 'FAKE_KEY'
                 }

    with pytest.raises(OSError):
        campaign = Campaigns(headers_1)
        campaign.create(campaign_data)

    with pytest.raises(OSError):
        campaign = Campaigns(headers_2)
        campaign.create(campaign_data)

    with pytest.raises(ValueError):
        campaign = Campaigns(headers_3)

    with pytest.raises(ValueError):
        campaign = Campaigns(headers_4)


def test_campaign_error(header):
    campaign = Campaigns(header)

    with pytest.raises(ValueError):
        campaign.count(status='inbox')

    with pytest.raises(IOError):
        campaign.all(order='inbox')

    with pytest.raises(ValueError):
        campaign.create(data=[("subject", "Regular campaign subject"),
                              ("type", "regular")])

    with pytest.raises(ValueError):
        campaign.create(data={"random_keys": "test"})

    with pytest.raises(ValueError):
        campaign.create(data={"random_keys": "test",
                              "type": "regular"})


def test_crud_campaign(header, campaign_data, campaign_data_ab):
    campaign_obj = Campaigns(header)

    code, res = campaign_obj.create(campaign_data)
    assert code == 200

    code, res = campaign_obj.create(campaign_data_ab)
    assert code == 200
    assert isinstance(res, dict)

    html = '<head></head><body><h1>Title</h1><p>Content</p><p><small>'
    html += '<a href=\"{$unsubscribe}\">Unsubscribe</a></small></p></body>'
    plain = "Your email client does not support HTML emails. "
    plain += "Open newsletter here: {$url}. If you do not want"
    plain += " to receive emails from us, click here: {$unsubscribe}"

    updated = campaign_obj.update(res['id'], html=html, plain=plain)
    assert updated
    assert isinstance(updated, bool)

    code, res = campaign_obj.delete(res['id'])
    assert code == 200
    assert res['success']

    res = campaign_obj.all(status='draft', limit=1000)
    nb_draft = campaign_obj.count('draft')
    assert nb_draft > 0
    assert len(res) > 0


def test_create_and_send_campaign(header, campaign_data):
    campaign_obj = Campaigns(header)

    code, res = campaign_obj.create(campaign_data)
    assert code == 200

    html = '<head></head><body><h1>Title</h1><p>Content</p><p><small>'
    html += '<a href=\"{$unsubscribe}\">Unsubscribe</a></small></p></body>'
    plain = "Your email client does not support HTML emails. "
    plain += "Open newsletter here: {$url}. If you do not want"
    plain += " to receive emails from us, click here: {$unsubscribe}"

    updated = campaign_obj.update(res['id'], html=html, plain=plain)

    code, res = campaign_obj.send(res['id'])
    assert code == 200


def test_cancel_send_campaign(header):
    campaign_obj = Campaigns(header)

    if campaign_obj.count('outbox'):
        res = campaign_obj.all(status='outbox', limit=5)
        assert res[0].status == 'outbox'
        code, res_2 = campaign_obj.cancel(res[0].id)
        assert code == 200
        assert res_2["status"] == 'draft'
        assert res[0].id == res_2["id"]



"""Utility function for calling the API."""

import requests
from urllib.parse import urlencode, urljoin

from mailerlite.constants import MAILERLITE_API_V2_URL, VALID_REQUEST_METHODS


def check_headers(headers):
    """Return True if the headers have the required keys.

    Parameters
    ----------
    headers : dict
        should contains 'content-type' and 'x-mailerlite-apikey'

    Returns
    -------
    valid_headers : bool
        True if it is valid
    error_msg : str
        the error message if there is any problem

    """
    valid_headers = True
    error_msg = ''
    if not headers:
        error_msg = "Empty headers. Please enter a valid one"
        valid_headers = False
        return valid_headers, error_msg

    if not isinstance(headers, dict):
        error_msg = "headers should be a dictionnary"
        valid_headers = False
        return valid_headers, error_msg

    if ('content-type' not in headers or
       'x-mailerlite-apikey' not in headers):
        error_msg = "headers should be a dictionnary and it contains "
        error_msg += "'content-type' and 'x-mailerlite-apikey' as keys"
        valid_headers = False
        return valid_headers, error_msg

    return valid_headers, error_msg


def build_url(*path, **queryparams):
    """Build path with endpoint and args.

    Parameters
    ----------
    path : endpoint url
    queryparams : dict

    Returns
    -------
    url : str
      desired path
    """
    url = '/'.join(map(str, path))
    if queryparams:
        url += '?{}'.format(urlencode(queryparams))
    return url


def make_request(url, method, headers=None, data=None,
                 timeout=None, hooks=None):
    """Make the request to the API.

    Parameters
    ----------
    url : str
    method : str
    headers : dict, optional
        Dictionary of HTTP Headers to send
    data : dict, optional
        A JSON serializable Python object to send in the body
    timeout : int, optional
        How long to wait for the server to send data before giving up
    hooks : dict, optional

    Returns
    -------
    response : int
        response value
    content : dict
        The JSON output from the API
    """
    if method not in VALID_REQUEST_METHODS:
        raise ValueError("Incorrect request method. method should be "
                         "{}".format(VALID_REQUEST_METHODS))

    url = urljoin(MAILERLITE_API_V2_URL, url)
    hooks = hooks or requests.hooks.default_hooks()
    headers = headers or requests.utils.default_headers()
    try:
        response = requests.request(**dict(method=method,
                                           url=url,
                                           json=data,
                                           timeout=timeout,
                                           hooks=hooks,
                                           headers=headers
                                           ))
    except requests.exceptions.RequestException as e:
        raise e
    else:
        if response.status_code >= 400:
            print(response.text)
            raise IOError(response)

        if response.status_code == 204:
            return None
        return response.status_code, response.json()

    return response.status_code, response.json()


def post(url, body=None, **kwargs):
    """Handle POST requests to add new information.

    Parameters
    ----------
    url : str
        The url for the endpoint including path parameters
    body : dict, optional
        The request body parameters. Default: None

    Returns
    -------
    response : int
        response value
    content : dict
        The JSON output from the API

    """
    return make_request(url=url, method='POST', data=body, **kwargs)


def get(url, params=None, **kwargs):
    """Handle GET requests to obtain information.

    Parameters
    ----------
    url: str
        The url for the endpoint including path parameters
    params: dict, optional
        The query string parameters

    Returns
    -------
    response : int
        response value
    content : dict
        The JSON output from the API
    """
    if params:
        url += '?' + urlencode(params)
    return make_request(url=url, method='GET', **kwargs)


def delete(url, **kwargs):
    """Handle DELETE requests to Remove information.

    Parameters
    ----------
    url: str
        The url for the endpoint including path parameters

    Returns
    -------
    response : int
        response value
    content : dict
        The JSON output from the API
    """
    return make_request(url=url, method='DELETE', **kwargs)


def put(url, body=None, **kwargs):
    """Handle PUT requests to modify existing information.

    Parameters
    ----------
    url : str
        The url for the endpoint including path parameters
    body : dict, optional
        The request body parameters. Default: None

    Returns
    -------
    response : int
        response value
    content : dict
        The JSON output from the API
    """
    return make_request(url=url, method='PUT', data=body, **kwargs)


def patch(url, body=None, **kwargs):
    """Handle PATCH requests.

    Parameters
    ----------
    url : str
        The url for the endpoint including path parameters
    body : dict, optional
        The request body parameters. Default: None

    Returns
    -------
    response : int
        response value
    content : dict
        The JSON output from the API
    """
    return make_request(url=url, method='PATCH', data=body, **kwargs)



"""Module to test subscriber class."""
import random
import string
import time

import pytest

from mailerlite.constants import API_KEY_TEST
from mailerlite.subscriber import Subscribers


@pytest.fixture
def header():
    headers = {'content-type': "application/json",
               'X-MailerLite-ApiDocs': "true",
               'x-mailerlite-apikey': API_KEY_TEST
               }
    return headers


def generate_random_email(length, seed=1234567):
    # random.seed(seed)
    letters = string.ascii_lowercase
    result_str = ''.join(random.choice(letters) for i in range(length))
    mail = 'demo-test-{}-{}@mailerlite.com'.format(result_str, seed)
    return mail


def test_wrong_headers():
    headers_2 = {'content-type': "application/json",
                 'x-mailerlite-apikey': 'FAKE_KEY'
                 }
    headers_3 = {'content-type': "application/json",
                 }
    headers_4 = {'x-mailerlite-apikey': 'FAKE_KEY'
                 }

    with pytest.raises(OSError):
        subs = Subscribers(headers_2)
        subs.all()

    with pytest.raises(ValueError):
        subs = Subscribers(headers_3)

    with pytest.raises(ValueError):
        subs = Subscribers(headers_4)


def test_subscribers_error(header):
    subs = Subscribers(header)

    # Unknown keys
    with pytest.raises(ValueError):
        subs.create("field_name")

    with pytest.raises(ValueError):
        subs.create({'mail': 'demo-test-12345555@mailerlite.com'})

    with pytest.raises(ValueError):
        subs.create({'email': 'demo-test-12345555@mailerlite.com',
                     'group': 12345
                     })

    with pytest.raises(OSError):
        subs.update(123456, "new_title")

    with pytest.raises(OSError):
        subs.get(mail='demo-test-12345555@mailerlite.com')


def test_subscribers_crud(header):
    subscriber = Subscribers(header)

    while True:
        try:
            num = random.randint(1000, 100000)
            mail = generate_random_email(length=15, seed=num)
            data = {'name': 'John',
                    'email': mail,
                    'fields': {'company': 'MailerLite'}
                    }
            e_res = subscriber.create(data)
        except OSError:
            time.sleep(3)
        else:
            break

    res = subscriber.get(email=mail)
    assert e_res.email == res.email

    res = subscriber.get(id=e_res.id)
    assert e_res.email == res.email

    res = subscriber.get(email=mail, as_json=True)
    assert e_res.email == res['email']

    res = subscriber.search(search=mail)
    assert len(res) > 0
    assert e_res.email == res[0].email

    with pytest.raises(IOError):
        subscriber.update({'email': 'new@mail.com'})

    with pytest.raises(IOError):
        subscriber.update({'email': 'new@mail.com'}, e_res.id)

    with pytest.raises(ValueError):
        subscriber.update({'email': 'new@mail.com'}, id=e_res.id)

    with pytest.raises(ValueError):
        subscriber.update({'random': 'new@mail.com'}, id=e_res.id)

    subscriber.update({'name': 'Jack'}, id=e_res.id)
    res = subscriber.get(id=e_res.id)
    assert e_res.name != res.name
    assert res.name == 'Jack'

    res = subscriber.active()
    assert len(res) > 0
    res = subscriber.unsubscribed()
    assert len(res) > 0
    res = subscriber.bounced()
    assert len(res) > 0
    res = subscriber.unconfirmed()
    assert len(res) > 0
    res = subscriber.junk()
    assert len(res) >= 0
    res = subscriber.all(as_json=True)
    assert len(res) > 0

    with pytest.warns(UserWarning):
        assert subscriber.count() == subscriber.count(
            as_json=True).get('count')

    groups = subscriber.groups(email=mail)
    assert len(groups) in [0, 1]

    activity = subscriber.activity(email=mail)
    assert len(activity) in [0, 1]

    with pytest.raises(IOError):
        subscriber.activity()

    with pytest.raises(ValueError):
        subscriber.activity(email=mail, atype='test')

    assert subscriber.delete(e_res.id) is None



"""Mailerlite API."""
import os

from mailerlite.campaign import Campaigns
from mailerlite.segment import Segments
from mailerlite.subscriber import Subscribers
from mailerlite.group import Groups
from mailerlite.field import Fields
from mailerlite.webhook import Webhooks
from mailerlite.account import Account
import mailerlite.client as client


class MailerLiteApi:
    """Python interface to the mailerlite v2 API.

    """

    def __init__(self, api_key=None):
        """Initialize a new mailerlite.api object.

        Parameters
        ----------
        api_key : str
            Your mailerlite api_key.

        """
        api_key = api_key or os.environ.get("MAILERLITE_PYTHON_API_KEY", None)

        if not api_key or not isinstance(api_key, str):
            raise ValueError("Empty API_KEY. Please enter a valid API_KEY")

        self._headers = {'content-type': "application/json",
                         "X-MailerLite-ApiDocs": "true",
                         'x-mailerlite-apikey': api_key
                         }

        self.campaigns = Campaigns(headers=self.headers)
        self.segments = Segments(headers=self.headers)
        self.subscribers = Subscribers(headers=self.headers)
        self.groups = Groups(headers=self.headers)
        self.fields = Fields(headers=self.headers)
        self.webhooks = Webhooks(headers=self.headers)
        self.account = Account(headers=self.headers)

    @property
    def headers(self):
        return self._headers

    def batch(self, batch_requests):
        """Execute a list of command. Dedicated for experts.

        https://developers.mailerlite.com/v2/reference#batch-api-requests

        Parameters
        ----------
        batch_requests : list of dict
            all you command. E.g:
            batch_requests = {"requests": [{"method":"GET",
                                            "path": "/api/v2/groups"
                                            },
                                            {"method":"POST",
                                             "path": "/api/v2/groups",
                                            "body": {"name": "New group"}
                                           }
                                          ]
                               }

        Returns
        -------
        results : list
            list of all desired object

        Notes
        -----
        * There is a limit of maximum 50 requests per single batch.
        * The order of response objects are the same as sent requests.
        * requests parameter should not be empty

        """
        url = client.build_url('batch')
        return client.post(url, body=batch_requests, headers=self.headers)



"""Module to test client."""
import mailerlite.client as client
from mailerlite.constants import API_KEY_TEST


def test_build_url():
    res = client.build_url('test', 125)
    assert res == 'test/125'

    res = client.build_url('test', 125, id='my_id')
    assert res == 'test/125?id=my_id'

    res = client.build_url('test', 125, id='my_id', value=123)
    assert res == 'test/125?id=my_id&value=123'


def test_check_headers():

    for headers in [{}, None, [], (), 0, '']:
        res, msg = client.check_headers(headers)
        assert not res
        assert "empty headers" in msg.lower()

    for headers in [['content-type'], ('content-type'), 1, 'content-type']:
        res, msg = client.check_headers(headers)
        assert not res
        assert "headers should be a dictionnary" in msg.lower()

    d1 = {'content-type': '', }
    d2 = {'x-mailerlite-apikey': '', }
    for headers in [d1, d2]:
        res, msg = client.check_headers(headers)
        assert not res
        assert "'content-type' and 'x-mailerlite-apikey'" in msg.lower()

    headers = {
        'content-type': "application/json",
        'X-MailerLite-ApiDocs': "true",
        'x-mailerlite-apikey': API_KEY_TEST
    }

    res, msg = client.check_headers(headers)
    assert res
    assert not msg



"""Module to tests groups."""
import random
import string
import time

import pytest

from mailerlite.constants import API_KEY_TEST, Group
from mailerlite.group import Groups


@pytest.fixture
def header():
    headers = {'content-type': "application/json",
               'X-MailerLite-ApiDocs': "true",
               'x-mailerlite-apikey': API_KEY_TEST
               }
    return headers


def generate_random_email(length, seed=1234567):
    # random.seed(seed)
    letters = string.ascii_lowercase
    result_str = ''.join(random.choice(letters) for i in range(length))
    mail = 'demo-test-{}-{}@mailerlite.com'.format(result_str, seed)
    return mail


def test_groups_instance(header):
    with pytest.raises(ValueError):
        Groups(API_KEY_TEST)

    groups = Groups(header)
    res_json = groups.all(as_json=True)

    # TEST to check if there is new key in the API
    missing_keys = []
    for g in res_json:
        for k in g.keys():
            if not hasattr(Group, k):
                missing_keys.append(k)

    missing_keys = set(missing_keys)
    error_msg = "The following keys are unkown: {}, ".format(missing_keys)
    error_msg += "please update API"
    assert not missing_keys, error_msg

    # Test to check if is group object is valid
    res = groups.all()
    assert len(res) > 0
    for g in res:
        assert isinstance(g, Group)


def test_groups_crud(header):
    groups = Groups(header)

    expected_group_name = "TEST_K_GROUP"
    expected_group_name_2 = "TEST_GROUP_KKK"
    e_res = groups.create(expected_group_name)
    assert e_res.name == expected_group_name

    res = groups.get(e_res.id)
    assert res.name == expected_group_name

    groups.update(e_res.id, expected_group_name_2)
    res = groups.get(e_res.id)
    assert res.name == expected_group_name_2

    assert groups.delete(e_res.id)

    with pytest.raises(OSError):
        groups.get(e_res.id)


def test_groups_subscriber(header):
    groups = Groups(header)

    n_groups = groups.all()
    assert len(n_groups) > 0
    group_1 = n_groups[0]

    subs_in_group_1 = groups.subscribers(group_1.id)
    assert len(subs_in_group_1) > 0

    sub1 = subs_in_group_1[0]
    tmp_sub = groups.subscriber(group_1.id, sub1.id)

    assert sub1.email == tmp_sub.email

    while True:
        try:
            num = random.randint(1000, 100000)
            mail = generate_random_email(length=15, seed=num)
            data = {'name': 'John',
                    'email': mail,
                    'fields': {'company': 'MailerLite'}
                    }
            new_subs = groups.add_subscribers(group_1.id, data)
        except OSError:
            time.sleep(3)
        else:
            break

    print(new_subs)
    if new_subs:
        assert new_subs[0].email == mail

        groups.delete_subscriber(group_1.id, new_subs[0].id)

    with pytest.raises(ValueError):
        data = {'name': 'John',
                'fields': {'company': 'MailerLite'}}
        groups.add_subscribers(group_1.id, data)

    with pytest.raises(ValueError):
        groups.add_subscribers(group_1.id, subscribers_data='hey!')


def test_groups_single_subscriber(header):
    groups = Groups(header)

    n_groups = groups.all()
    assert len(n_groups) > 0
    group_1 = n_groups[0]

    subs_in_group_1 = groups.subscribers(group_1.id)
    assert len(subs_in_group_1) > 0

    sub1 = subs_in_group_1[0]
    tmp_sub = groups.subscriber(group_1.id, sub1.id)

    assert sub1.email == tmp_sub.email

    while True:
        try:
            num = random.randint(1000, 100000)
            mail = generate_random_email(length=15, seed=num)
            data = {'name': 'John',
                    'email': mail,
                    'fields': {'company': 'MailerLite'}
                    }
            new_sub = groups.add_single_subscriber(group_1.id, data)
        except OSError:
            time.sleep(3)
        else:
            break

    print(new_sub)
    if new_sub:
        assert new_sub.email == mail

        groups.delete_subscriber(group_1.id, new_sub.id)

    with pytest.raises(ValueError):
        data = {'name': 'John',
                'fields': {'company': 'MailerLite'}}
        groups.add_single_subscriber(group_1.id, data)



"""Manage Webhooks."""

import mailerlite.client as client
from mailerlite.constants import Webhook


class Webhooks:

    def __init__(self, headers):
        """Initialize Webhooks object.

        Parameters
        ----------
        headers : dict
            request header containing your mailerlite api_key.
            More information : https://developers.mailerlite.com/docs/request
        """
        valid_headers, error_msg = client.check_headers(headers)
        if not valid_headers:
            raise ValueError(error_msg)

        self.headers = headers

    def all(self, as_json=False):
        """Get list of Webhooks.

        https://developers.mailerlite.com/v2/reference#get-webhooks-list

        Returns
        -------
        webhooks: list of dict
            all webhooks.
        as_json : bool
            return result as json format

        """
        url = client.build_url('webhooks')
        _, res_json = client.get(url, headers=self.headers)

        if as_json or not res_json:
            return res_json

        all_webhooks = [Webhook(**res) for res in res_json.get('webhooks')]
        return all_webhooks

    def get(self, webhook_id, as_json=False):
        """Get single field by ID from your account.

        https://developers.mailerlite.com/v2/reference#get-single-webhook

        Parameters
        ----------
        webhook_id : int
            ID of a webhook
        as_json : bool
            return result as json format

        Returns
        -------
        webhook: dict
            the desired webhook.

        """
        url = client.build_url('webhooks', webhook_id)
        _, res_json = client.get(url, headers=self.headers)

        if as_json or not res_json:
            return res_json

        webhook = Webhook(**res_json)
        return webhook

    def delete(self, webhook_id):
        """Remove a webhook.

        https://developers.mailerlite.com/v2/reference#delete-a-webhook

        Parameters
        ----------
        webhook_id : int
            ID of a webhook

        Returns
        -------
        success: bool
            deletion status
        """
        url = client.build_url('webhooks', webhook_id)
        return client.delete(url, headers=self.headers)

    def update(self, webhook_id, url, event):
        """Update webhook.

        https://developers.mailerlite.com/v2/reference#update-a-webhook

        Parameters
        ----------
        webhook_id : int
            ID of a webhook
        url : str
            Your URL where callbacks are sent
        event : str
            Subscribed event

        Returns
        -------
        webhook : :class:Webhook
            webhook object updated
        """
        url = client.build_url('webhooks', webhook_id)
        body = {"url": url, 'event': event}
        _, res_json = client.put(url, body=body, headers=self.headers)

        return res_json

    def create(self, url, event):
        """Create a webhook.

        https://developers.mailerlite.com/v2/reference#create-a-webhook

        Parameters
        ----------
        url : str
            Your URL where callbacks are sent
        event : str
            Subscribed event

        Returns
        -------
        field : :class:Field
            field object updated
        """
        url = client.build_url('webhooks')
        body = {"url": url, 'event': event}
        code, res_json = client.post(url, body=body, headers=self.headers)

        webhook = Webhook(**res_json)
        return code, webhook

    def count(self):
        """Return the number of webhooks.

        Returns
        -------
        count: int
            number of webhooks

        """
        res_json = self.all(as_json=True)
        return res_json.get('count') or len(res_json.get('webhooks'))



#!/usr/bin/env python
"""Simple tools to query github.com and gather repo information.

Taken from ipython

"""
# ----------------------------------------------------------------------------
# Imports
# ----------------------------------------------------------------------------
from distutils.version import LooseVersion

import json
import re
import sys
import os
import argparse
import operator

from datetime import datetime, timedelta
from subprocess import check_output

from urllib.request import urlopen

# ----------------------------------------------------------------------------
# Globals
# ----------------------------------------------------------------------------

ISO8601 = "%Y-%m-%dT%H:%M:%SZ"
PER_PAGE = 100

element_pat = re.compile(r'<(.+?)>')
rel_pat = re.compile(r'rel=[\'"](\w+)[\'"]')

LAST_RELEASE = datetime(2015, 3, 18)
CONTRIBUTORS_FILE = "contributors.json"
GH_TOKEN = os.environ.get('GH_TOKEN', '')
TOKEN_URL = "access_token={}".format(GH_TOKEN) if GH_TOKEN else ''


# ----------------------------------------------------------------------------
# Functions
# ----------------------------------------------------------------------------


def fetch_url(url):
    if TOKEN_URL:
        if "?" in url:
            url += "&{0}".format(TOKEN_URL)
        else:
            url += "?{0}".format(TOKEN_URL)
    try:
        print("fetching %s" % url, file=sys.stderr)
        # url = Request(url,
        #               headers={'Accept': 'application/vnd.github.v3+json',
        #                        'User-agent': 'Defined'})
        f = urlopen(url)
    except Exception as e:
        print(e)
        print("return Empty data", file=sys.stderr)
        return {}

    return f


def parse_link_header(headers):
    link_s = headers.get('link', '')
    urls = element_pat.findall(link_s)
    rels = rel_pat.findall(link_s)
    d = {}
    for rel, url in zip(rels, urls):
        d[rel] = url
    return d


def get_json_from_url(url):
    """Fetch and read url."""
    f = fetch_url(url)
    return json.load(f) if f else {}


def get_paged_request(url):
    """Get a full list, handling APIv3's paging."""
    results = []
    while url:
        f = fetch_url(url)
        if not f:
            continue
        results.extend(json.load(f))
        links = parse_link_header(f.headers)
        url = links.get('next')
    return results


def get_issues(project="skoudoro/mailerlite-api-python", state="closed",
               pulls=False):
    """Get a list of the issues from the Github API."""
    which = 'pulls' if pulls else 'issues'
    url = "https://api.github.com/repos/%s/%s?state=%s&per_page=%i" % \
        (project, which, state, PER_PAGE)
    return get_paged_request(url)


def get_tags(project="skoudoro/mailerlite-api-python"):
    """Get a list of the tags from the Github API."""
    url = "https://api.github.com/repos/{0}/tags".format(project)
    return get_paged_request(url)


def fetch_basic_stats(project="skoudoro/mailerlite-api-python"):
    """Fetch the basic stats.

    Returns
    -------
    basic_stats : dict
        A dictionary containing basic statistics. For example:
        {'subscribers': 41,
         'forks': 142,
         'forks_url': 'https://github.com/skoudoro/mailerlite-api-python/network'
         'watchers': 94,
         'open_issues': 154,
         'stars': 94,
         'stars_url': 'https://github.com/skoudoro/mailerlite-api-python/stargazers'
        }

    """
    desired_keys = ["stargazers_count", "stargazers_url", "watchers_count",
                    "watchers_url", "forks_count", "forks_url", "open_issues",
                    "issues_url", "subscribers_count", "subscribers_url"]
    url = "https://api.github.com/repos/{0}".format(project)
    r_json = get_json_from_url(url)
    basic_stats = dict((k, r_json[k]) for k in desired_keys if k in r_json)
    return basic_stats


def fetch_contributor_stats(project="skoudoro/mailerlite-api-python"):
    """Fetch stats of contributors.

    Returns
    -------
    contributor_stats : dict
        A dictionary containing contributor statistics. For example:
        {'total_contributors': 50,
         'total_commits': 6031,
         'contributors': [ {
            "user_name": "Garyfallidis"
            "avatar_url":"https://avatars.githubusercontent.com/u/134276?v=3",
            "html_url": "https://github.com/Garyfallidis",
            "total_commits": 1389,
            "total_additions": 116712,
            "total_deletions": 70340,
            "weekly_commits": [
                        {
                        "w": "1367712000",
                        "a": 6898,
                        "d": 77,
                        "c": 10
                        },
                    ]
            },
            ]
        }

    """
    url = "https://api.github.com/repos/{0}/stats/contributors".format(project)
    r_json = get_json_from_url(url)

    contributor_stats = {}
    contributor_stats["total_contributors"] = len(r_json)
    contributor_stats["contributors"] = []

    cumulative_commits = 0
    desired_keys = ["login", "avatar_url", "html_url"]
    with open(os.path.join(os.path.dirname(__file__), "..",
                           CONTRIBUTORS_FILE)) as f:
        extra_info = json.load(f)
        extra_info = extra_info["team"] + extra_info["core_team"]

    for contributor in r_json:
        contributor_dict = dict((k, contributor["author"][k])
                                for k in desired_keys
                                if k in contributor["author"])

        # check if "author" is null
        if not contributor_dict["login"]:
            continue

        # Replace key name
        contributor_dict["username"] = usrname = contributor_dict.pop("login")
        contributor_dict["nb_commits"] = contributor["total"]

        # Add extra information like fullname and affiliation
        l_extra_info = [e for e in extra_info
                        if e["username"].lower() == usrname.lower()]

        l_extra_info = l_extra_info[0] if l_extra_info else {}
        contributor_dict.update(l_extra_info)

        # Update total commits
        cumulative_commits += contributor_dict["nb_commits"]

        total_additions = 0
        total_deletions = 0
        for week in contributor["weeks"]:
            total_additions += week['a']
            total_deletions += week['d']

        contributor_dict["total_additions"] = total_additions
        contributor_dict["total_deletions"] = total_deletions
        # contributor_dict["weekly_commits"] = contributor["weeks"]
        contributor_stats["contributors"].insert(0, contributor_dict)

    contributor_stats["total_commits"] = cumulative_commits
    return contributor_stats


def _parse_datetime(s):
    """Parse dates in the format returned by the Github API."""
    if s:
        return datetime.strptime(s, ISO8601)
    else:
        return datetime.fromtimestamp(0)


def issues2dict(issues):
    """Convert a list of issues to a dict, keyed by issue number."""
    idict = {}
    for i in issues:
        idict[i['number']] = i
    return idict


def is_pull_request(issue):
    """Return True if the given issue is a pull request."""
    return 'pull_request_url' in issue


def issues_closed_since(period=LAST_RELEASE,
                        project="skoudoro/mailerlite-api-python",
                        pulls=False):
    """Get all issues closed since a particular point in time.

    period can either be a datetime object, or a timedelta object. In the
    latter case, it is used as a time before the present.

    """
    which = 'pulls' if pulls else 'issues'

    if isinstance(period, timedelta):
        period = datetime.now() - period
    url = ("https://api.github.com/repos/%s/%s?state=closed&sort=updated&"
           "since=%s&per_page=%i") % (project, which,
                                      period.strftime(ISO8601), PER_PAGE)
    allclosed = get_paged_request(url)
    # allclosed = get_issues(project=project, state='closed', pulls=pulls,
    #                        since=period)
    filtered = [i for i in allclosed
                if _parse_datetime(i['closed_at']) > period]

    # exclude rejected PRs
    if pulls:
        filtered = [pr for pr in filtered if pr['merged_at']]

    return filtered


def sorted_by_field(issues, field='closed_at', reverse=False):
    """Return a list of issues sorted by closing date date."""
    return sorted(issues, key=lambda i: i[field], reverse=reverse)


def report(issues, show_urls=False):
    """Summary report about a list of issues, printing number and title."""
    # titles may have unicode in them, so we must encode everything below
    if show_urls:
        for i in issues:
            role = 'ghpull' if 'merged_at' in i else 'ghissue'
            print('* :%s:`%d`: %s' % (role, i['number'],
                                      i['title']))
    else:
        for i in issues:
            print('* %d: %s' % (i['number'], i['title']))


def get_all_versions(ignore='', project="skoudoro/mailerlite-api-python"):
    """Return all releases version.

    Parameters
    ----------
    ignore: str
        skip a version number between (default: '')
        you can skip minor, or micro version number, it
        will be replace by x
    project: str
        repo path

    Returns
    -------
    l_version: list of str
        versions number list

    """
    tags = get_tags(project=project)
    l_version = [t['name'] for t in tags]

    if ignore.lower() in ['micro', 'minor']:
        l_version = list(set([re.sub(r'(\d+)$', 'x', v) for v in l_version]))

    if ignore.lower() == 'minor':
        l_version = list(set([re.sub(r'\.(\d+)\.', '.x.', v)
                              for v in l_version]))

    return l_version


def version_compare(current_version, version_number, op='eq',
                    all_versions=None):
    """Compare doc version. This is afilter for sphinx template."""
    p = re.compile(r'(\d+)\.(\d+)')
    d_operator = {'<': operator.lt,
                  'lt': operator.lt,
                  '<=': operator.le,
                  'le': operator.le,
                  '>': operator.gt,
                  'gt': operator.gt,
                  '>=': operator.ge,
                  'ge': operator.ge,
                  '==': operator.eq,
                  '=': operator.eq,
                  'eq': operator.eq,
                  '!=': operator.ne,
                  }
    # Setup default value to op
    if op not in d_operator.keys():
        op = 'eq'

    # check dev page
    if current_version.lower() == 'dev':
        return 'post' in version_number

    # major and minor extraction
    current = p.search(current_version)
    ref = p.search(version_number)

    if current_version.lower() == 'latest':
        # Check if it is the latest release
        all_versions = all_versions or get_all_versions()
        if not all_versions:
            return False
        last_version = sorted(all_versions)[0]
        last_version = p.search(last_version)
        if LooseVersion(last_version.group()) ==  \
           LooseVersion(ref.group()) and \
           'post' not in version_number:
            return True
        return False

    if 'post' in version_number:
        return False

    return d_operator[op](LooseVersion(current.group()),
                          LooseVersion(ref.group()))


def username_to_fullname(all_authors):
    contributors_file = os.path.join(os.path.dirname(__file__), "..",
                                     CONTRIBUTORS_FILE)

    if not os.path.isfile(contributors_file):
        return all_authors

    with open(contributors_file) as f:
        extra_info = json.load(f)
        extra_info = extra_info["team"] + extra_info["core_team"]
        extra_info = {i["username"]: i['fullname'] for i in extra_info}

    curent_authors = all_authors
    for i, author in enumerate(all_authors):
        if author[2:] in extra_info.keys():
            curent_authors[i] = '* ' + extra_info[author[2:]]

    return curent_authors


def github_stats(**kwargs):
    """Get release github stats."""
    # Whether to add reST urls for all issues in printout.
    show_urls = True
    save = kwargs.get('save', None)
    if save:
        version = kwargs.get('version', 'vx.x.x')
        fname = "releasev" + version + '.rst'
        fpath = os.path.join(os.path.dirname(__file__), '..', 'release_notes',
                             fname)
        f = open(fpath, 'w')
        orig_stdout = sys.stdout
        sys.stdout = f
        print(".. _{}".format(fname))
        print()
        print(("==============================\n"
               " Release notes version {}\n"
               "==============================").format(version))

    # By default, search one month back
    tag = kwargs.get('tag', None)
    if tag is None:
        if len(sys.argv) > 1:
            try:
                days = int(sys.argv[1])
            except Exception:
                tag = sys.argv[1]
        else:
            tag = check_output(['git', 'describe', '--abbrev=0'],
                               universal_newlines=True).strip()

    if tag:
        cmd = ['git', 'log', '-1', '--format=%ai', tag]
        tagday, _ = check_output(cmd,
                                 universal_newlines=True).strip().rsplit(' ',
                                                                         1)
        since = datetime.strptime(tagday, "%Y-%m-%d %H:%M:%S")
    else:
        since = datetime.now() - timedelta(days=days)

    print("fetching GitHub stats since %s (tag: %s)" % (since, tag),
          file=sys.stderr)

    issues = issues_closed_since(since, pulls=False)
    pulls = issues_closed_since(since, pulls=True)

    # For regular reports, it's nice to show them in reverse chronological
    # order
    issues = sorted_by_field(issues, reverse=True)
    pulls = sorted_by_field(pulls, reverse=True)

    n_issues, n_pulls = map(len, (issues, pulls))
    n_total = n_issues + n_pulls

    # Print summary report we can directly include into release notes.
    print()
    since_day = since.strftime("%Y/%m/%d")
    today = datetime.today().strftime("%Y/%m/%d")
    print("GitHub stats for %s - %s (tag: %s)" % (since_day, today, tag))
    print()
    print("These lists are automatically generated, and may be incomplete or"
          " contain duplicates.")
    print()
    if tag:
        # print git info, in addition to GitHub info:
        since_tag = tag+'..'
        cmd = ['git', 'log', '--oneline', since_tag]
        ncommits = check_output(cmd, universal_newlines=True).splitlines()
        ncommits = len(ncommits)

        author_cmd = ['git', 'log', '--format=* %aN', since_tag]
        all_authors = check_output(author_cmd, universal_newlines=True) \
            .splitlines()
        # Replace username by author name
        all_authors = username_to_fullname(all_authors)
        unique_authors = sorted(set(all_authors))

        if len(unique_authors) == 0:
            print("No commits during this period.")
        else:
            print("The following %i authors contributed %i commits." %
                  (len(unique_authors), ncommits))
            print()
            print('\n'.join(unique_authors))
            print()

            print()
            print("We closed a total of %d issues, %d pull requests and %d"
                  " regular issues;\nthis is the full list (generated with"
                  " the script \n:file:`tools/github_stats.py`):" %
                  (n_total, n_pulls, n_issues))
            print()
            print('Pull Requests (%d):\n' % n_pulls)
            report(pulls, show_urls)
            print()
            print('Issues (%d):\n' % n_issues)
            report(issues, show_urls)

    if save:
        sys.stdout = orig_stdout
        f.close()


# ----------------------------------------------------------------------------
# Sphinx connection
# ----------------------------------------------------------------------------
def add_jinja_filters(app):
    app.builder.templates.environment.filters['version_compare'] = \
        version_compare


def setup(app):
    """
    - Create releases information
    - Collect and clean authors
    - Adds extra jinja filters.
    """
    app.connect("builder-inited", add_jinja_filters)
    app.add_stylesheet("css/custom_github.css")


# ----------------------------------------------------------------------------
# Main script
# ----------------------------------------------------------------------------


if __name__ == "__main__":
    # e.g github_tools.py --tag=v0.1.3 --save --version=0.1.4
    parser = argparse.ArgumentParser()
    parser.add_argument("--tag", type=str, default=None,
                        help='from which tag version to get information')
    parser.add_argument("--version", type=str, default='',
                        help='current release version')
    parser.add_argument("--save", dest='save', action='store_true',
                        default=False, help=("Save in the release folder"
                                             "and add rst header")
                        )

    args = parser.parse_args()
    github_stats(**vars(args))



"""Manage Fields."""

import mailerlite.client as client
from mailerlite.constants import Field


class Fields:

    def __init__(self, headers):
        """Initialize Fields object.

        Parameters
        ----------
        headers : dict
            request header containing your mailerlite api_key.
            More information : https://developers.mailerlite.com/docs/request

        """
        valid_headers, error_msg = client.check_headers(headers)
        if not valid_headers:
            raise ValueError(error_msg)

        self.headers = headers

    def all(self, as_json=False):
        """Get list of fields from your account.

        https://developers.mailerlite.com/v2/reference#all-fields

        Parameters
        ----------
        as_json : bool
            return result as json format

        Returns
        -------
        fields: list
            all desired Fields.

        """
        url = client.build_url('fields')
        _, res_json = client.get(url, headers=self.headers)

        if as_json or not res_json:
            return res_json

        all_fields = [Field(**res) for res in res_json]
        return all_fields

    def get(self, field_id, as_json=False):
        """Get single field by ID from your account.

        look at https://developers.mailerlite.com/v2/reference#all-fields

        Parameters
        ----------
        field_id : int
            should be group id. e.g: id=1343965485
        as_json : bool
            return result as json format

        Returns
        -------
        Field: :class:Field
            a single field

        Notes
        -----
        The endpoint GET /fields/id do not exist, so we trick it.

        """
        all_fields = self.all()
        desired_field = [field for field in all_fields
                         if field.id == field_id]
        desired_field = desired_field[0] if desired_field else None

        return desired_field

    def delete(self, field_id):
        """Remove custom field from account.

        https://developers.mailerlite.com/v2/reference#remove-field

        Parameters
        ----------
        field_id : int
            field id

        Returns
        -------
        success: bool
            deletion status

        """
        url = client.build_url('fields', field_id)
        return client.delete(url, headers=self.headers)

    def update(self, field_id, title, as_json=False):
        """Update custom field in account.

        https://developers.mailerlite.com/v2/reference#update-field

        Parameters
        ----------
        field_id : int
            field id
        title : str
            Title of field

        Returns
        -------
        field : :class:Field
            field object updated

        """
        url = client.build_url('fields', field_id)
        body = {"title": title}
        _, res_json = client.put(url, body=body, headers=self.headers)

        if as_json or not res_json:
            return res_json

        return Field(**res_json)

    def create(self, title, field_type='TEXT'):
        """Create new custom field in account.

        https://developers.mailerlite.com/v2/reference#create-field

        Parameters
        ----------
        title : str
            Title of field
        field_type: str
            Type of field. Available values: TEXT , NUMBER, DATE
            (default: TEXT)

        Returns
        -------
        field : :class:Field
            field object updated

        """
        if field_type.upper() not in ['TEXT', 'NUMBER', 'DATE']:
            raise ValueError('Incorrect field_type. Available values'
                             ' are: TEXT , NUMBER, DATE')
        url = client.build_url('fields')
        data = {'title': title, 'type': field_type.upper()}
        return client.post(url, body=data, headers=self.headers)



"""Module to tests Field."""
import random
import string

import pytest

from mailerlite.constants import API_KEY_TEST, Field
from mailerlite.field import Fields


@pytest.fixture
def header():
    headers = {'content-type': "application/json",
               'X-MailerLite-ApiDocs': "true",
               'x-mailerlite-apikey': API_KEY_TEST
               }
    return headers


def generate_random_string(length, seed=1234567):
    # random.seed(seed)
    letters = string.ascii_uppercase
    result_str = ''.join(random.choice(letters) for i in range(length))
    return result_str


def test_wrong_headers():
    headers_1 = {'content-type': "app",
                 'x-mailerlite-apikey': API_KEY_TEST
                 }
    headers_2 = {'content-type': "application/json",
                 'x-mailerlite-apikey': 'FAKE_KEY'
                 }
    headers_3 = {'content-type': "application/json",
                 }
    headers_4 = {'x-mailerlite-apikey': 'FAKE_KEY'
                 }

    with pytest.raises(OSError):
        field = Fields(headers_1)
        field.create("field_new", "TEXT")

    with pytest.raises(OSError):
        field = Fields(headers_2)
        field.create("field_new", "TEXT")

    with pytest.raises(ValueError):
        field = Fields(headers_3)

    with pytest.raises(ValueError):
        field = Fields(headers_4)


def test_fields_error(header):
    fields = Fields(header)

    # Unknown keys
    with pytest.raises(ValueError):
        fields.create("field_name", "UNKNOW_FIELD")

    with pytest.raises(OSError):
        fields.update(123456, "new_title")


def test_fields_crud(header):
    fields = Fields(header)
    all_fields = fields.all()
    assert len(all_fields) > 0

    last_field = all_fields[-2]
    current_field = fields.get(last_field.id)

    assert isinstance(current_field, Field)
    for f in current_field._fields:
        assert current_field._asdict().get(f) == last_field._asdict().get(f)

    field_name = "TEST_"
    field_name += generate_random_string(random.randint(5, 15))
    code, custom_fields = fields.create(field_name, 'TEXT')
    assert code in [200, 201]

    custom_fields = Field(**custom_fields)
    assert custom_fields.title == field_name
    assert custom_fields.key == field_name.lower()
    assert custom_fields.type == 'TEXT'

    updated = fields.update(custom_fields.id, "TEST_MAILERLITE_PYTHON")
    assert updated.title == 'TEST_MAILERLITE_PYTHON'
    assert updated.key == custom_fields.key
    assert updated.type == custom_fields.type

    assert fields.delete(updated.id) is None
    assert fields.get(updated.id) is None
